/**
 * @file seriala.c
 * The AmigaOS serial line handler
 * @author Marko Mkel <marko.makela at iki.fi>
 */

/* Copyright  2001,2002,2004,2006 Marko Mkel.

   This file is part of C2N, a program for processing data tapes in
   Commodore C2N format and other formats.

   C2N is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   C2N is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   The GNU General Public License is often shipped with GNU software, and
   is generally kept in a file called COPYING or LICENSE.  If you do not
   have a copy of the license, write to the Free Software Foundation,
   59 Temple Place, Suite 330, Boston, MA 02111 USA. */

#include "c2n.h"
#include "serial.h"
#ifdef HAS_SERIAL
/** The name of the serial device */
static const char* serialdev;
# include <stdio.h>
# include <string.h>
# include <errno.h>
# include <proto/exec.h>
# include <devices/serial.h>
# include <devices/timer.h>
# include <dos/dos.h>
# include <clib/alib_protos.h>
# include <signal.h>

# ifdef SERIAL_DEBUG
/* log file for data received from the serial line */
static FILE* serial_debug;
# endif /* SERIAL_DEBUG */
/** extended input/output request structure */
struct IOExtSer* ser;
/** timer request */
static struct timerequest* timer;
/** flag: read or write pending */
static int pending = 0;
/** Report an error for accessing the serial device
 * @param op	the failed operation
 */
static void
report_error (const char* op)
{
  fprintf (stderr, "c2n -c %s: %s returned %d\n",
	   serialdev, op, ser->IOSer.io_Error);
}
/** Wait for an operation on the serial line to complete, or optional timeout
 * @param tout	timeout interval (optional)
 * @return	the WaitIO exit status; pending=0, or 1 if timeout occurred
 */
static BYTE
wait_serial (struct timerequest* tout)
{
  const long waitsigs = SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D |
    (tout ? 1L << tout->tr_node.io_Message.mn_ReplyPort->mp_SigBit : 0) |
    1L << ser->IOSer.io_Message.mn_ReplyPort->mp_SigBit;
  if (tout) {
    tout->tr_node.io_Command = TR_ADDREQUEST;
    SendIO ((struct IORequest*) tout);
  }
  pending = 1;
  for (;;) {
    if (Wait (waitsigs) & (SIGBREAKF_CTRL_C | SIGBREAKF_CTRL_D)) {
      if (tout) {
	AbortIO ((struct IORequest*) tout);
	WaitIO ((struct IORequest*) tout);
      }
      AbortIO ((struct IORequest*) ser);
      WaitIO ((struct IORequest*) ser);
      pending = 0;
      raise (SIGINT);
    }
    if (CheckIO ((struct IORequest*) ser)) {
      pending = 0;
      if (tout) {
	/* serial operation completed: stop the timer */
	AbortIO ((struct IORequest*) tout);
	WaitIO ((struct IORequest*) tout);
      }
      return WaitIO ((struct IORequest*) ser);
    }
    else if (tout && CheckIO ((struct IORequest*) tout)) {
      /* timeout occurred: caller stops the serial operation */
#ifdef SERIAL_DEBUG
      fputs ("\ntimeout\n", serial_debug);
#endif /* SERIAL_DEBUG */
      WaitIO ((struct IORequest*) tout);
      return 0;
    }
  }
}

/** input/output buffer */
static char buf[8192];
/** second output buffer */
static char buf2[sizeof buf];
/** pointer to the active output buffer */
static char* actbuf = buf;
/** number of bytes read from or written to buf */
static unsigned datain;

/** Switch output buffers */
static void
switch_buf (void)
{
  if (pending)
    wait_serial (0);
  if (datain) {
    pending = 1;
    ser->IOSer.io_Command = CMD_WRITE;
    ser->IOSer.io_Length = datain;
    ser->IOSer.io_Data = actbuf;
    SendIO ((struct IORequest*) ser);
    actbuf = actbuf != buf ? buf : buf2;
    datain = 0;
  }
}

/** Set up the serial communications
 * @param dev		serial device driver name
 * @param fast		nonzero=fast mode (unsafe with USB RS-232 adapters)
 * @return		zero on success, nonzero on failure
 */
int
serial_init (const char* dev, int fast)
{
  struct MsgPort* port;
#ifdef SERIAL_DEBUG
  if (!serial_debug && !(serial_debug = fopen ("serial.debug", "wb"))) {
    perror ("fopen(\"serial.debug\")");
    return 1;
  }
#endif /* SERIAL_DEBUG */
  /* open serial.device */
  if (!(port = CreatePort (0, 0))) {
    fputs ("CreatePort for " SERIALNAME " failed\n", stderr);
    return 1;
  }
  if (!(ser = (struct IOExtSer*)
	CreateExtIO (port, sizeof *ser))) {
    DeletePort (port);
    fputs ("CreateExtIO for IOExtSer failed\n", stderr);
    return 2;
  }
  ser->io_SerFlags = 0;
  ser->IOSer.io_Device = 0;
  serialdev = dev;
  if (OpenDevice ((const unsigned char*) serialdev, 0,
		  (struct IORequest*) ser, 0)) {
    report_error ("OpenDevice");
    return 3;
  }
  /* open timer.device */
  if (!(port = CreatePort (0, 0))) {
    fputs ("CreatePort for " TIMERNAME " failed\n", stderr);
    return 4;
  }
  if (!(timer = (struct timerequest*)
	CreateExtIO (port, sizeof *timer))) {
    DeletePort (port);
    fputs ("CreateExtIO for timerequest failed\n", stderr);
    return 5;
  }
  timer->tr_node.io_Device = 0;
  if (OpenDevice ((const unsigned char*) TIMERNAME,
		  UNIT_VBLANK, (struct IORequest*) timer, 0L)) {
    fprintf (stderr, "OpenDevice for " TIMERNAME " returned %d\n",
	     timer->tr_node.io_Error);
    return 6;
  }

  ser->IOSer.io_Command = SDCMD_SETPARAMS;
  ser->io_Baud = 38400;
  ser->io_ExtFlags = 0;
  ser->io_RBufLen = 2 * sizeof buf;
  ser->io_ReadLen = ser->io_WriteLen = 8;
  ser->io_StopBits = 1;
  if (DoIO ((struct IORequest*) ser)) {
    report_error ("DoIO (SDCMD_SETPARAMS)");
    return 2;
  }

  /* send a NUL character and await a response */
  ser->IOSer.io_Command = CMD_WRITE;
  ser->IOSer.io_Length = 1;
  ser->IOSer.io_Data = buf;
  *buf = 0;
  SendIO ((struct IORequest*) ser);
  if (wait_serial (0)) {
    report_error ("WaitIO (CMD_WRITE)");
    return 2;
  }
  ser->IOSer.io_Command = CMD_READ;
  ser->IOSer.io_Length = sizeof buf;
  ser->IOSer.io_Data = buf;
  SendIO ((struct IORequest*) ser);
  timer->tr_time.tv_secs = 0;
  timer->tr_time.tv_micro = 500000L;
  if (wait_serial (timer)) {
    report_error ("WaitIO (CMD_READ)");
    return 2;
  }
  if (pending) {
    pending = 0;
    AbortIO ((struct IORequest*) ser);
    WaitIO ((struct IORequest*) ser);
  }
  if (!ser->IOSer.io_Actual) {
    fputs ("-c ", stderr);
    fputs (serialdev, stderr);
    fputs (": no response from C2N232 device\n", stderr);
    return 2;
  }
#ifdef SERIAL_DEBUG
  fwrite (buf, 1, ser->IOSer.io_Actual, serial_debug);
#endif /* SERIAL_DEBUG */
  if (buf[ser->IOSer.io_Actual - 1] != '0') {
    fputs ("-c ", stderr);
    fputs (serialdev, stderr);
    fputs (": unexpected response from C2N232 device\n", stderr);
    return 2;
  }
  return 0;
}

void
serial_cleanup (void)
{
  struct MsgPort* port;
#ifdef SERIAL_DEBUG
  if (serial_debug)
    fclose (serial_debug), serial_debug = 0;
#endif /* SERIAL_DEBUG */
  if (ser) {
    port = ser->IOSer.io_Message.mn_ReplyPort;
    if (ser->IOSer.io_Device) {
      ser->IOSer.io_Command = CMD_CLEAR;
      if (DoIO ((struct IORequest*) ser))
	fprintf (stderr, "cleanup: DoIO (CMD_CLEAR) returned %d\n",
		 ser->IOSer.io_Error);
      else {
	ser->IOSer.io_Command = SDCMD_BREAK;
	if (DoIO ((struct IORequest*) ser))
	  fprintf (stderr, "cleanup: DoIO (SDCMD_BREAK) returned %d\n",
		   ser->IOSer.io_Error);
      }
      CloseDevice ((struct IORequest*) ser);
    }
    DeleteExtIO ((struct IORequest*) ser);
    ser = 0;
    DeletePort (port);
  }
  if (timer) {
    port = timer->tr_node.io_Message.mn_ReplyPort;
    if (timer->tr_node.io_Device)
      CloseDevice ((struct IORequest*) timer);
    DeleteExtIO ((struct IORequest*) timer);
    timer = 0;
    DeletePort (port);
  }
}

/** Read a pulse from the serial line
 * @return	the read pulse
 */
enum pulse
sp_read (void)
{
  char c;
  if (!serial_open)
    return Pause;
 again:
  if (datain < ser->IOSer.io_Actual)
    c = buf[datain++];
  else {
    if (pending)
      goto pending;
    datain = 0;
    ser->IOSer.io_Command = CMD_READ;
    ser->IOSer.io_Length = sizeof buf;
    ser->IOSer.io_Data = buf;
    SendIO ((struct IORequest*) ser);
#ifdef SERIAL_DEBUG
    fprintf (serial_debug, "\nsp_read %lu:", ser->IOSer.io_Actual);
    fwrite (buf, 1, ser->IOSer.io_Actual, serial_debug);
#endif /* SERIAL_DEBUG */
  pending:
    timer->tr_time.tv_secs = 2;
    timer->tr_time.tv_micro = 0;
    if (wait_serial (timer)) {
      report_error ("sp_read: WaitIO (CMD_READ)");
      goto cleanup;
    }
    if (pending && datain == ser->IOSer.io_Actual) {
      AbortIO ((struct IORequest*) ser);
      WaitIO ((struct IORequest*) ser);
      pending = 0;
      cleanup:
	serial_cleanup ();
	return Pause;
    }
    goto again;
  }
  switch (c) {
  case Pause: case Short: case Medium: case Long:
    return c;
  }
  fprintf (stderr, "unrecognized character 0x%02x in pulse input stream\n",
	   (unsigned char) c);
  goto again;
}

/** Empty the serial buffers */
void
serial_empty (void)
{
  if (!pending)
    datain = 0, ser->IOSer.io_Actual = 0;
}

/** Flush the serial output buffer */
void
serial_flush (void)
{
  switch_buf ();
  switch_buf ();
}

/** Send a command to the C2N232 device
 * @param buf	the command block
 * @param len	the length of the block
 * @return	zero on success, nonzero on failure
 */
int
serial_command (unsigned char* buf,
		unsigned len)
{
  ser->IOSer.io_Command = CMD_WRITE;
  ser->IOSer.io_Length = len;
  ser->IOSer.io_Data = buf;
  if (DoIO ((struct IORequest*) ser)) {
    report_error ("DoIO (CMD_WRITE)");
    return 1;
  }
  ser->IOSer.io_Command = CMD_READ;
  ser->IOSer.io_Length = 1;
  ser->IOSer.io_Data = buf + 1;
  if (DoIO ((struct IORequest*) ser)) {
    report_error ("DoIO (CMD_READ)");
    return 2;
  }

  if ((buf[1] & ~0x30) == buf[0])
    return 0;
  fprintf (stderr, "C2N232 responded 0x%02x; expected 0x%02x\n",
	   buf[1], buf[0] | 0x30);
  return 3;
}

/** Disable the serial read timeout
 * @return	zero on success, nonzero on failure
 */
int
serial_timeout (void)
{
  ser->IOSer.io_Command = SDCMD_SETPARAMS;
  ser->io_SerFlags = 0;
  if (DoIO ((struct IORequest*) ser)) {
    report_error ("DoIO (SDCMD_SETPARAMS)");
    return 1;
  }
  return 0;
}

/** Make the C2N232 go idle */
void
serial_go_idle (void)
{
  if (pending)
    wait_serial (0);
  ser->IOSer.io_Command = CMD_WRITE;
  ser->IOSer.io_Length = 1;
  ser->IOSer.io_Data = buf;
  *buf = 0;
  DoIO ((struct IORequest*) ser);
}

/** Write a pulse to the serial line or to standard output
 * @param p	the pulse to be written
 */
void
p_write (enum pulse p)
{
  if (serial_open) {
    actbuf[datain++] = p;
    if (p == Pause || datain == sizeof buf)
      switch_buf ();
  }
  else
    putchar (p);
}
#endif /* HAS_SERIAL */
